/*
   lex.c -- PIKT common lex routines

   $Header: /var/cvs/PIKT/pikt/src/lib/lex.c,v 1.24 2007/01/18 13:13:09 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2007 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org
 */

#include "common.h"

/* return whether the leading parenthesis in the supplied string is a
   logic conditional parenthesis, as opposed to an arithmetic parenthesis;
   basically, we count the number of left versus right parentheses, until
   we reach certain turning points, where if the number of counted left
   parentheses exceeds the right count, we know that we are entering a
   logical condition
 */
int
condparen (char *s)
{
	register int i;
	register int lp, rp, dqt;
	register int escape;
	register unsigned char c;
	unsigned char nc;
	int done;
	int context;
	int wasspace;

	done = FALSE;
	context = NIL;
	wasspace = FALSE;
	escape = FALSE;
	lp = rp = dqt = 0;
	i = 0;
	while ((c = *(s+i))) {
		switch(c) {
			case '\\':
				if (! escape) {
					escape = TRUE;
				} else {
					escape = FALSE;
				}
				break;
			case '\"':
				if (! escape) {
					dqt++;
					context = STRING;
				}
				escape = FALSE;
				break;
			case '$':
				if (! escape) {
					if (context != NUMBER)
						context = STRING;
				}
				escape = FALSE;
				break;
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':
			case '#':
				if (! escape) {
					if (even(dqt)) {
						if (context != STRING)
							context = NUMBER;
					}
				}
				escape = FALSE;
				break;
			case LP:
				if (! escape) {
					if (even(dqt))
						lp++;
				}
				escape = FALSE;
				break;
			case RP:
				if (! escape) {
					if (even(dqt)) {
						rp++;
						if (lp == rp) {
							if (context == STRING)
								rp--;	/* signalling cond */
							done = TRUE;
						}
					}
				}
				escape = FALSE;
				break;
			case '<':
				if (! escape) {
					if (even(dqt)) {
						nc = *(s+i+1);
						switch(nc) {
							case ' ':
							case '>':
							case '\n':
							case '\t':
								done = TRUE;
								break;
							default:
								break;
						}
					}
				}
				escape = FALSE;
				break;
			case '>':
				if (! escape) {
					if (even(dqt)) {
						nc = *(s+i+1);
						switch(nc) {
							case ' ':
							case '\n':
							case '\t':
								done = TRUE;
								break;
							default:
								break;
						}
					}
				}
				escape = FALSE;
				break;
			case '!':
				if (! escape) {
					if (even(dqt)) {
						nc = *(s+i+1);
						if (isspace((int)nc) && (lp == 1))
							done = TRUE;
					}
				}
				escape = FALSE;
				break;
			case '-':	/* e.g., (-e $name) */
				if (! escape) {
					if (even(dqt)) {
						nc = *(s+i+1);
						if (strchr("ezfdcbplS",nc) && (lp == 1))
							done = TRUE;
					}
				}
				escape = FALSE;
				break;
			case 'e':
				if (! escape) {
					if (even(dqt)) {
						if (wasspace) {
							if (! (nc = *(s+i+1)))
								break;
							switch(nc) {
								case 'q':
									nc = *(s+i+2);
									if (strchr(" \t\n", nc))
										done = TRUE;
									break;
								default:
									break;
							}
						}
					}
				}
				escape = FALSE;
				break;
			case 'n':
				if (! escape) {
					if (even(dqt)) {
						if (wasspace) {
							if (! (nc = *(s+i+1)))
								break;
							switch(nc) {
								case 'e':
									nc = *(s+i+2);
									if (strchr(" \t\n", nc))
										done = TRUE;
									break;
								default:
									break;
							}
						}
					}
				}
				escape = FALSE;
				break;
			case 'g':
			case 'l':
				if (! escape) {
					if (even(dqt)) {
						if (wasspace) {
							if (! (nc = *(s+i+1)))
								break;
							switch(nc) {
								case 'e':
								case 't':
									nc = *(s+i+2);
									if (strchr(" \t\n", nc))
										done = TRUE;
									break;
								default:
									break;
							}
						}
					}
				}
				escape = FALSE;
				break;
			default:
				escape = FALSE;
				break;
		}
		if (done) break;
		if (isspace((int)c))
			wasspace = TRUE;
		else
			wasspace = FALSE;
		i++;
	}
	return lp > rp ? TRUE : FALSE;
}

/* when processing a statement like

      $substitute($line, "t", "X", 2) eq "this is a Xest"

   the lexer will eventually pattern match

      "t", "X", 2) eq "this is a Xest"

   this function isolates the next function string, i.e.,

      "t"

   returning its length, in this example, 3
 */
int
leadstr (char *s)
{
	register int i;
	register unsigned char c;
	register int brkt, prn, dqt;
	register int escape;

	escape = FALSE;
	brkt = prn = dqt = 0;
	i = 0;
	while ((c = *(s+i++))) {
		if ((c == '\\') && ! escape) {
			escape = TRUE;
			continue;
		} else if ((c == LB) && ! escape) {
			brkt++;
		} else if ((c == RB) && ! escape) {
			brkt--;
		} else if ((c == LP) && ! escape) {
			prn++;
		} else if ((c == RP) && ! escape) {
			prn--;
		} else if ((c == '\"') && ! escape) {
			dqt++;
			if (even(dqt) && ! brkt && ! prn)
				break;
		}
		escape = FALSE;
	}
	return i;
}

/* given a string, break it down into its component parts--pure text strings,
   surrounded by single quotes, and variables or functions, joined together
   by the concatenation operator . (dot);

   for example, the input string

      "User $uname has uid $uid > the 60000 upper limit"

   when decomposed becomes

      'User ' . $uname . ' has uid ' . $uid . ' > the 60000 upper limit'

   the function returns the length of the output string minus 1

   the decomposition output string is pushed back into the lexer input
   stream for subsequent reprocessing
 */
int
decompose (signed char *istr, signed char *ostr)
{
	static char fnnam[] = "decompose";
	register signed char *cp;
	register int i;
	register signed char c;
	int isvar;
	register int escape;
	register int brkt, prn, dqt;

	cp = istr;
	isvar = FALSE;
	escape = FALSE;
	brkt = prn = dqt = 0;
	i = 0;
	while ((c = *cp)) {
		if ((c == '\\') && ! escape)  {
			/*
			if (odd(dqt))
				*(ostr+i++) = c;
			*/
			if (isvar) {
				*(ostr+i) = '\0';
				if (odd(dqt)) {
					strcat(ostr, " . \'");
					i += 4;
				}
			}
			isvar = FALSE;
			escape = TRUE;
			cp++;
			continue;
		}
		if (isspace((int)c) && (brkt || prn) && even(dqt)) {
			*(ostr+i++) = c;
			isvar = FALSE;
			escape = FALSE;
			cp++;
			continue;
		}
		if (isvar) {
			if ((c == LB) && ! prn && ! escape) {
				brkt++;
				*(ostr+i++) = c;
				escape = FALSE;
				cp++;
				continue;
			} else if ((c == RB) && ! prn && brkt && ! escape) {
				brkt--;
				*(ostr+i++) = c;
				escape = FALSE;
				cp++;
				continue;
			} else if ((c == LP) && ! brkt && ! escape) {
				prn++;
				*(ostr+i++) = c;
				escape = FALSE;
				cp++;
				continue;
			} else if ((c == RP) && ! brkt && prn && ! escape) {
				prn--;
				*(ostr+i++) = c;
				if (! prn) {
					*(ostr+i) = '\0';
					if (odd(dqt)) {
						strcat(ostr, " . \'");
						i += 4;
					}
					isvar = FALSE;
				}
				escape = FALSE;
				cp++;
				continue;
			} else if (   isalnum((int)c)
				   || (((brkt > 0) && ! prn) || ((prn > 0) && ! brkt))
				  ) {
				if ((c == '\"') && ! escape) {
					*(ostr+i++) = '\'';
					dqt++;
				} else if (c == '\"') {
					if (*(cp+1)) {
						*(ostr+i++) = DQTPROXY;
					} else {
						*(ostr+i++) = '\'';
						dqt++;
					}
				} else if (c == '\'') {
					*(ostr+i++) = SQTPROXY;
				} else {
					*(ostr+i++) = c;
				}
				escape = FALSE;
				cp++;
				continue;
			} else {
				if (ISVAR(c)) {
					*(ostr+i) = '\0';
					strcat(ostr, " . ");
					i += 3;
					*(ostr+i++) = c;
					escape = FALSE;
					cp++;
					continue;
				} else {
					*(ostr+i) = '\0';
					if (odd(dqt)) {
						strcat(ostr, " . \'");
						i += 4;
					}
					isvar = FALSE;
				}
			}
		} else { 
			if (ISVAR(c)) {
				if ((cp != istr+1) || ((cp == istr+1) && (c != '$'))) {
					*(ostr+i) = '\0';
					if (odd(dqt) && i) {
						strcat(ostr, "\' . ");
						i += 4;
					}
				}
				*(ostr+i++) = c;
				isvar = TRUE;
				escape = FALSE;
				cp++;
				continue;
			}
		}
		if ((c == '\"') && ! escape) {
			if ((cp == istr) && *(cp+1) && strchr("$#%@",*(cp+1)) && (isalnum((int)*(cp+2)) || (*(cp+2) == '\['))) {
				dqt++;	/* bypass opening quote if
					   immediately followd by a var,
					   but be sure to count it!
					 */
			} else {
				*(ostr+i++) = '\'';
				dqt++;
			}
		} else if (c == '\"') {
			if (*(cp+1)) {
				*(ostr+i++) = DQTPROXY;
			} else {
				*(ostr+i++) = '\'';
				dqt++;
			}
		} else if (c == '\'') {
			*(ostr+i++) = SQTPROXY;
		} else {
			*(ostr+i++) = c;
		}
		escape = FALSE;
		cp++;
	}
	*(ostr+i) = '\0';
	/* the following if is for the case where the istr is, e.g.,
	   "\$trim", and where regtok would (mistakenly) detect an
	   attempt to use a reserved function name as a var name;
	 */
	if (*(istr+1) == '\\') {
		if (strchr("$#%@",*(ostr+1)) && ! strncmp(istr+2, ostr+1, (size_t)(i-2))) {
			return 0;
		}
	}
	if (isdebug)
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "istr is %s, ostr is %s, i is %d", istr, ostr, i);
	return i-1;
}

/* rewrite the given #exec or #piktexec line, returning the rewritten line's
   length
 */
int
getexecunput (char **eu, char *yyt, int yylen, int ispexec, int isdevnull, int isexpand)
{
	static char fnnam[] = "getexecunput";
	FILE *infile;
	char buf[MAXBUF+1];
	int buflen;
	register int j;
	char *stm;

	/* get the <proc> */
	j = yylen;
	/* retreat to ']' */
	while (*(yyt+j) != ']')
		j--;
	*(yyt+j) = '\0';
	j = 0;
	/* advance to '[' */
	while (*(yyt+j) != '[')
		j++;
	/* move beyond '[' */
	j++;

	if (ispexec) {
		/* if is a #piktexec directive, possibly expand macros with
		   a call to 'piktc -M +M'
		 */
		if (! isexpand) {
			stm = mycalloc((unsigned)(strlen(yyt+j)+strlen(piktc)+strlen(hostname)+14), sizeof(char), fnnam);
			sprintf(stm, "%s -M +M '%s' +H %s", piktc, yyt+j, hostname);
			if ((infile = popen(stm, "r"))) {
				if (fgets(buf, MAXBUF, infile) == NULL) {
					if (ferror(infile))
						logreturn(PIKTLOG, LOG_ERR, ERRID_FGETS_ERROR, __FILE__, __LINE__, fnnam, FALSE, "fgets() error for %s", stm);
					strcpy(buf, "quit");
					buflen = 4;
				} else {
					*strchr(buf, '\n') = '\0';
					buflen = strlen(buf);
				}
			        pclose(infile);
			} else {
				logreturn(PIKTLOG, LOG_ERR, ERRID_PROC_OPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "popen() failure for %s", stm);
				strcpy(buf, "quit");
				buflen = 4;
			}
			free(stm);
		} else {
			strcpy(buf, "quit");
			buflen = 4;
		}
	} else {
		/* else no macro expansion, take the <proc> as is */
		strcpy(buf, yyt+j);
		buflen = strlen(buf);
	}

	if (ispexec) {
		if (isdevnull) {
			*eu = mycalloc((unsigned)(strlen(pikt)+buflen+31), sizeof(char), fnnam);
			sprintf(*eu, "#include \"/dev/null\" [%s +C '%s']\n", pikt, buf);
		} else {
			*eu = mycalloc((unsigned)(strlen(pikt)+buflen+35), sizeof(char), fnnam);
			sprintf(*eu, "#include [%s +C '%s' >/dev/null 2>&1]\n", pikt, buf);
		}
	} else {
		if (isdevnull) {
			*eu = mycalloc((unsigned)(buflen+25), sizeof(char), fnnam);
			sprintf(*eu, "#include \"/dev/null\" [%s]\n", buf);
		} else {
			*eu = mycalloc((unsigned)(buflen+29), sizeof(char), fnnam);
			sprintf(*eu, "#include [%s >/dev/null 2>&1]\n", buf);
		}
	}

	return strlen(*eu);
}
