/*
   statement.c -- PIKT alarm script statement routines

   $Header: /var/cvs/PIKT/pikt/src/piktd/statement.c,v 1.69 2008/05/13 17:14:58 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2007 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org
*/

#include "../lib/common.h"

#include "piktd.h"

/* perform various actions for current program statement; in simple
   cases (such as REDO), just return with a value indicating where next
   (what statement) to go in program (quite often to immediate next
   program statement (as with ENDIF, for example); in complex cases,
   current statement has expressions to be evaluated; in the case of
   setting an array variable, an expression on left-hand side of "="
   needs to be evaluated; in most other cases (and also in array setting
   case), a right-hand side expression must be evaluated; after expression
   evaluation, perform various actions appropriate to current statement,
   using evaluated expressions; in all cases, return a value indicating
   where next (what statement) to go in program flow for, example TRUE
   (for tnext statement), FALSE (for fnext statement), AGAIN, LEAVE, REDO,
   NEXT, LAST, QUIT, etc.
 */
int
dostatement (int type)
{
	static char fnnam[] = "dostatement";
	char *cmd;
	pid_t pid;
	vl *vlp;
	register int k;

	if (istest) {
		if ((type == EXEC) || (type == EXECWAIT)) {
			if (interactive ) {
				type = OUTPUT;
			} else {
				type = MAIL;
			}
		}
	}
	switch (type) {
		case AGAIN:	/* apply current rule again */
			return AGAIN;
			break;
		case LEAVE:	/* leave current rule and go on to next */
			return LEAVE;
			break;
		case REDO:	/* redo current line */
			return REDO;
			break;
		case NEXT:	/* go on to next line */
			return NEXT;
			break;
		case LAST:	/* go to end */
			return LAST;
			break;
		case QUIT:	/* go on to next alarm */
			return QUIT;
			break;
		case ELSE:
		case ENDIF:
		case ENDWHILE:
		case WHILEAUX:
		case REPEAT:
		case REPEATAUX:
		case ENDFOR:
		case FORAUX:
		case SWITCH:
		case DEFAULT:
		case ENDSWITCH:
			return TRUE;
			break;
		case BREAK:
		case CONTINUE:
		case BREAKSWITCH:
			return FALSE;
			break;
		case SET:
		case UNSET:
		case LOCLOG:
			if (stmcurr->lhsexpr) {
				tokcurr = stmcurr->lhsexpr;
				exprres = NIL;
				exprval = NIL;
				exprparse();
				if (! exprstr) {
					logreturn(PIKTLOG, LOG_ERR, ERRID_EXPRSTR_NULL, __FILE__, __LINE__, fnnam, FALSE, "in %s, null exprstr", ALMNAM);
					STATUS = ERROR;
					return QUIT;
				}
				if (type == LOCLOG) {
					stmcurr->logfile = mystrdup(exprstr, fnnam);
				} else {
					if (exprval != ERRARNDX) {
						for (k=0; k<=MAXARYDIM; k++)
							stmcurr->arndx[k] = arndx[k];
						FREE(stmcurr->arkey);
						stmcurr->arkey = mystrdup(exprstr, fnnam);
					} else {
						logreturn(PIKTLOG, LOG_ERR, ERRID_ARRAY_INDEX_INVALID, __FILE__, __LINE__, fnnam, FALSE, "in %s, bad array index: %f", ALMNAM, exprval);
						for (k=0; k<=MAXARYDIM; k++)
							stmcurr->arndx[k] = ERRARNDX;
						FREEVOID(stmcurr->arkey);
						return TRUE;
					}
				}
			}
			/* no break here! */
		default:	/* all other cases have rhsexprs */
			tokcurr = stmcurr->rhsexpr;
			break;
	}
	if (exprres == TRUE) exprres = NIL;
	exprval = NIL;
	exprparse();
	if (! exprstr) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_EXPRSTR_NULL, __FILE__, __LINE__, fnnam, FALSE, "in %s, null exprstr", ALMNAM);
		STATUS = ERROR;
		return QUIT;
	}
	requote(exprstr);
	switch (type) {
		case DO:
			return TRUE;
			break;
		case IF:
		case ELIF:
		case WHILE:
		case UNTIL:
		case FOR:
		case CASE:
			if (exprres == TRUE)
				return TRUE;
			else
				return FALSE;
			break;
		case SKIP:	/* go ahead n lines */
			skip = (int) exprval;
			return SKIP;
			break;
		case PAUSE:
			if ((int)exprval >= 0)
				sleep((unsigned int)exprval);
			else
				logreturn(PIKTLOG, LOG_ERR, ERRID_VALUE_ZERO_OR_NEGATIVE, __FILE__, __LINE__, fnnam, FALSE, "in %s, illegal (negative or zero) pause argument: %f", ALMNAM, exprval);
			return TRUE;
			break;
		case DIE:	/* exit program */
			if (stmcurr->status == ERROR) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_STATEMENT_ERROR, __FILE__, __LINE__, fnnam, FALSE, "in %s, bad die line: %s", ALMNAM, exprstr);
			}
			if (isautosyslog)
				logreturn(SYSLOG, LOG_ERR, ERRID_DIE_STATEMENT, __FILE__, __LINE__, fnnam, FALSE, "in %s, %s", ALMNAM, exprstr);
			logexit(PIKTLOG, LOG_ERR, ERRID_DIE_STATEMENT, __FILE__, __LINE__, fnnam, FALSE, "in %s, %s", ALMNAM, exprstr);
			break;
		case OUTPUT:
		case MAIL:
			if (stmcurr->status == ERROR) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_STATEMENT_ERROR, __FILE__, __LINE__, fnnam, FALSE, "in %s, line not output, error in line: %s", ALMNAM, exprstr);
				stmcurr->status = ACTIVE;
				return TRUE;
			} else {
				if (! interactive && (type != OUTPUT)) {
					if (newmailalert) {
						altheader(&mailhdrmsg, &mailhdrmsgcurr, type);
						newmailalert = FALSE;
					}
					if (newmailalarm) {
						almheader(&mailalmmsg[issortmessages ? LEVEL : 0], &mailalmmsgcurr[issortmessages ? LEVEL : 0], type);
						newmailalarm = FALSE;
					}
				}
				almline(&mailalmmsg[issortmessages ? LEVEL : 0], &mailalmmsgcurr[issortmessages ? LEVEL : 0], exprstr, type);
				if (isautosyslog && (type != OUTPUT))
					logreturn(SYSLOG, LEVEL, ERRID_SYSLOG_ENTRY, __FILE__, __LINE__, fnnam, FALSE, "%s", exprstr);
				return TRUE;
			}
			break;
		case PRINT:
			if (stmcurr->status == ERROR) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_STATEMENT_ERROR, __FILE__, __LINE__, fnnam, FALSE, "in %s, line not output, error in line: %s", ALMNAM, exprstr);
				stmcurr->status = ACTIVE;
				return TRUE;
			} else {
				if (! interactive) {
					if (newprintalert) {
						altheader(&printhdrmsg, &printhdrmsgcurr, type);
						newprintalert = FALSE;
					}
					if (newprintalarm) {
						almheader(&printalmmsg[issortmessages ? LEVEL : 0], &printalmmsgcurr[issortmessages ? LEVEL : 0], type);
						newprintalarm = FALSE;
					}
				}
				almline(&printalmmsg[issortmessages ? LEVEL : 0], &printalmmsgcurr[issortmessages ? LEVEL : 0], exprstr, type);
				if (isautosyslog)
					logreturn(SYSLOG, LEVEL, ERRID_SYSLOG_ENTRY, __FILE__, __LINE__, fnnam, FALSE, "%s", exprstr);
				return TRUE;
			}
			break;
		case LOCLOG:
			if (stmcurr->status == ERROR) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_STATEMENT_ERROR, __FILE__, __LINE__, fnnam, FALSE, "in %s, bad log line: %s", ALMNAM, exprstr);
				stmcurr->status = ACTIVE;
			}
			if (strcmp(stmcurr->logfile, "/dev/null")) {
				logreturn(LOCLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "%s", exprstr);
				/* but what if write failure to LOCLOG? */
				if (isautosyslog)
					logreturn(SYSLOG, LEVEL, ERRID_SYSLOG_ENTRY, __FILE__, __LINE__, fnnam, FALSE, "%s", exprstr);
			}
			return TRUE;
			break;
		case PIKTLOG:
			if (stmcurr->status == ERROR) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_STATEMENT_ERROR, __FILE__, __LINE__, fnnam, FALSE, "in %s, bad log line: %s", ALMNAM, exprstr);
				stmcurr->status = ACTIVE;
			}
			logreturn(PIKTLOG, LEVEL, ERRID_PIKTLOG_ENTRY, __FILE__, __LINE__, fnnam, FALSE, "%s", exprstr);
			if (isautosyslog)
				logreturn(SYSLOG, LEVEL, ERRID_SYSLOG_ENTRY, __FILE__, __LINE__, fnnam, FALSE, "%s", exprstr);
			return TRUE;
			break;
		case SYSLOG:
			if (stmcurr->status == ERROR) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_STATEMENT_ERROR, __FILE__, __LINE__, fnnam, FALSE, "in %s, bad log line: %s", ALMNAM, exprstr);
				stmcurr->status = ACTIVE;
			}
			logreturn(SYSLOG, LEVEL, ERRID_SYSLOG_ENTRY, __FILE__, __LINE__, fnnam, FALSE, "%s", exprstr);
			return TRUE;
			break;
		case SET:
			if (INPUTTMPFIL && (stmcurr->vrndx == OUTLINNOW)) {
				REWRITEINFIL = TRUE;
				outlinnow = mystrdup(exprstr, fnnam);
				/* FREEVOID(stmcurr->arkey); */
				return TRUE;
			}
			if (VALTYP(stmcurr->vrndx) & FIXEDVAL) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_VAR_SET_ATTEMPTED, __FILE__, __LINE__, fnnam, FALSE, "in %s, attempt to set data variable %s", ALMNAM, VARNAM(stmcurr->vrndx));
				STATUS = ERROR;
				FREEVOID(stmcurr->arkey);
				return QUIT;
			}
			if (VALTYP(stmcurr->vrndx) & BEGINVAL) {
				if ((vlp = lookupvl(valprc, stmcurr->vrndx, stmcurr->arndx, stmcurr->arkey, ARYTYP(stmcurr->vrndx)))) {
					FREE(vlp->str);
				} else {
					if (! (vlp = regvl(valprc, stmcurr->vrndx, stmcurr->arndx, stmcurr->arkey, ARYTYP(stmcurr->vrndx)))) {
						logreturn(PIKTLOG, LOG_ERR, ERRID_REGVL_ERROR, __FILE__, __LINE__, fnnam, FALSE, "in %s, regvl() error for var %s", ALMNAM, VARNAM(stmcurr->vrndx));
						STATUS = ERROR;
						FREEVOID(stmcurr->arkey);
						return QUIT;
					}
				}
				vlp->str = mystrdup(exprstr, fnnam);
				vlp->num = exprval;
				/* should be unnecessary:
				vlp->fil = (FILE *)NULL;
				 */
			}
			if (VALTYP(stmcurr->vrndx) & (BEGINVAL|FREEVAL|ENDVAL)) {
				if ((vlp = lookupvl(valnow, stmcurr->vrndx, stmcurr->arndx, stmcurr->arkey, ARYTYP(stmcurr->vrndx)))) {
					FREE(vlp->str);
				} else {
					if (! (vlp = regvl(valnow, stmcurr->vrndx, stmcurr->arndx, stmcurr->arkey, ARYTYP(stmcurr->vrndx)))) {
						logreturn(PIKTLOG, LOG_ERR, ERRID_REGVL_ERROR, __FILE__, __LINE__, fnnam, FALSE, "in %s, regvl() error for var %s", ALMNAM, VARNAM(stmcurr->vrndx));
						STATUS = ERROR;
						FREEVOID(stmcurr->arkey);
						return QUIT;
					}
				}
				vlp->str = mystrdup(exprstr, fnnam);
				vlp->num = exprval;
				/* should be unnecessary:
				vlp->fil = (FILE *)NULL;
				 */
				MAPKEYSVRNDXARY(stmcurr->vrndx) = NIL;
			}
			if (isdebug) {
				vlp = lookupvl(valnow, stmcurr->vrndx, stmcurr->arndx, stmcurr->arkey, ARYTYP(stmcurr->vrndx));
				if (vlp && vlp->str) {
					printf("DEBUG: %s (str) set to: %s\n", VARNAM(stmcurr->vrndx), vlp->str);
					fflush(stdout);
				}
				if (vlp && (vlp->num != NIL)) {
					printf("DEBUG: %s (num) set to: %f\n", VARNAM(stmcurr->vrndx), vlp->num);
					fflush(stdout);
				}
			}
			FREEVOID(stmcurr->arkey);
			return TRUE;
			break;
		case UNSET:
			/*
			if (VALTYP(stmcurr->vrndx) & FIXEDVAL) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_VAR_UNSET_ATTEMPTED, __FILE__, __LINE__, fnnam, FALSE, "in %s, attempt to unset data variable %s", ALMNAM, VARNAM(stmcurr->vrndx));
				STATUS = ERROR;
				FREEVOID(stmcurr->arkey);
				return QUIT;
			}
			*/
			if (VALTYP(stmcurr->vrndx) & BEGINVAL) {
				delvl(valprc, stmcurr->vrndx, stmcurr->arndx, stmcurr->arkey, ARYTYP(stmcurr->vrndx));
			}
			if (VALTYP(stmcurr->vrndx) & (BEGINVAL|FREEVAL|ENDVAL)) {
				delvl(valnow, stmcurr->vrndx, stmcurr->arndx, stmcurr->arkey, ARYTYP(stmcurr->vrndx));
			}
			if (isdebug) {
				printf("DEBUG: %s[%s] unset\n", VARNAM(stmcurr->vrndx), stmcurr->arkey ? stmcurr->arkey : nullstring);
				fflush(stdout);
			}
			delarkey(stmcurr->vrndx, &(ARKEYS(stmcurr->vrndx)), stmcurr->arkey, FALSE);
			delarkey(stmcurr->vrndx, &(SARKEYS(stmcurr->vrndx)), stmcurr->arkey, TRUE);
			FREEVOID(stmcurr->arkey);
			return TRUE;
			break;
		case EXEC:
		case DOEXEC:
			if (stmcurr->status == ERROR) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_STATEMENT_INACTIVE, __FILE__, __LINE__, fnnam, FALSE, "in %s, command not exec'ed, error in command: %s", ALMNAM, exprstr);
				STATUS = ERROR;
				return QUIT;
			} else {
				if (! *exprstr) {
					logreturn(PIKTLOG, LOG_ERR, ERRID_EXPRSTR_NULL, __FILE__, __LINE__, fnnam, FALSE, "in %s, null statement string", ALMNAM);
					return TRUE;
				}
				switch (pid = fork()) {
					case -1:	/* error */
						logexit(PIKTLOG, LOG_ERR, ERRID_FORK_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, fork() error", ALMNAM);
						break;
					case 0:		/* child */
					     	/* replace the parent's process id with the child's */
						prgpid = getpid();
						cmd = (char *) mymalloc(strlen(exprstr)+strlen(piktlog)+6, fnnam);
						sprintf(cmd, "%s 2>> %s", exprstr, piktlog);
						if (isautosyslog)
							logreturn(SYSLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "in %s, %s \"%s\"", ALMNAM, type == EXEC ? "exec" : "doexec", cmd);
						logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "in %s, %s \"%s\"", ALMNAM, type == EXEC ? "exec" : "doexec", cmd);
						execl(SHELL, "sh", "-c", cmd, (char *) NULL);
						free(cmd);
						_exit(127);
						break;
					default:	/* parent */
						break;
				}
				return TRUE;
			}
			break;
		case EXECWAIT:
		case DOEXECWAIT:
			if (stmcurr->status == ERROR) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_STATEMENT_INACTIVE, __FILE__, __LINE__, fnnam, FALSE, "in %s, command not exec'ed, error in command: %s", ALMNAM, exprstr);
				STATUS = ERROR;
				return QUIT;
			} else {
				if (! *exprstr) {
					logreturn(PIKTLOG, LOG_ERR, ERRID_STATEMENT_INACTIVE, __FILE__, __LINE__, fnnam, FALSE, "in %s, null statement string", ALMNAM);
					return TRUE;
				}
				/* oh boy, more lock files!  i am reluctant
				   to do this, since lock file management
				   can get messy, but i see no other way;
				   we don't check for a lock file here,
				   because we don't want an orphaned lock
				   file from an aborted previous run to block,
				   or result in error; rather, we only block
				   below, after the fork, when we know we may
				   have a legitimate lock file; note that
				   any orphaned lock files are removed after
				   each alarm completion; for debugging
				   purposes, we now write the exprstr to the
				   lock file
				 */
				if (mklock(execlockfile, exprstr, interactive, TRUE) == ERROR) {
					logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_CREATE_ERROR, __FILE__, __LINE__, fnnam, FALSE, "in %s, mklock() error for %s", ALMNAM, execlockfile);
					/* no return: if we can't make the
					   lock file, we'll just treat this
					   like a simple EXEC (no wait)
					return TRUE;
					 */
				}
				switch (pid = fork()) {
					case -1:	/* error */
						logexit(PIKTLOG, LOG_ERR, ERRID_FORK_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, fork() error", ALMNAM);
						break;
					case 0:		/* child */
					     	/* replace the parent's process id with the child's */
						prgpid = getpid();
						cmd = (char *) mymalloc(strlen(exprstr)+strlen(piktlog)+strlen(rmcmd)+strlen(execlockfile)+24, fnnam);
						sprintf(cmd, "%s 2>> %s; %s -f %s 2>/dev/null", exprstr, piktlog, rmcmd, execlockfile);
						if (isautosyslog)
							logreturn(SYSLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "in %s, %s wait \"%s\"", ALMNAM, type == EXECWAIT ? "exec" : "doexec", cmd);
						logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "in %s, %s wait \"%s\"", ALMNAM, type == EXECWAIT ? "exec" : "doexec", cmd);
						/* exec the cmd, then
						   remove the lock
						 */
						execl(SHELL, "sh", "-c", cmd, (char *) NULL);
						free(cmd);
						_exit(127);
						break;
					default:	/* parent */
						/* wait until the child
						   removes the lock
						 */
						while (islock(execlockfile)) {
							sleep(EXECWAITWAIT);
						}
						break;
				}
				return TRUE;
			}
			break;
	}
	return QUIT;
}
